/*
 * wmrotozoom is a windowmaker/afterstep dock/wharf applet for GNU/Linux
 *
 * It draws a warped version of the area around your cursor.
 *
 * Code based on some of the Window maker programs found on the internet
 *
 * Code based on wmpong, by Dag Wieers <dag@digibel.be>
 *
 * Copyright 1999, Dave Turner (turnerd@reed.edu);
 *
 * This program is distributed under the GPL license. 
 *
 * Thanks to Richard Stallman for the GNU system and the GPL.
 * 
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <math.h>

#include <X11/xpm.h>

#include "../wmgeneral/wmgeneral.h"

#include "wmrotozoom_master.xpm"

#define VERSION  "0.1"

#define SPEED 3
#define X_CENTER 32
#define Y_CENTER 32
#define SCALE_DISTANCE_AMT 1.2
#define SCALE_POSITION_AMT 0.8
#define X_DATA_FORMAT int

#define USLEEP 20000

#ifndef PI
#define PI 3.1415926535897932384626
#endif


struct xpm_position
  {
    int x, y;
  };
struct xpm_surface
  {
    int x, y, w, h;
  };



struct xpm_position pos[] =
{
  {5, 50},
  {16, 50},
  {27, 50},
  {38, 50},
  {49, 50}
};
struct xpm_surface rel[] =
{
  {1, 70, 11, 9},
  {12, 70, 11, 9},
  {23, 70, 11, 9},
  {34, 70, 11, 9},
  {45, 70, 11, 9}
};
struct xpm_surface prs[] =
{
  {1, 79, 11, 9},
  {12, 79, 11, 9},
  {23, 79, 11, 9},
  {34, 79, 11, 9},
  {45, 79, 11, 9}
};

int screen_width, screen_height;


extern GC NormalGC;

float spherewidth [64];
float tx [64];

/*****************************************************************************/
void init_sphere () {

  int i;

  for (i = 0; i < 64; i ++) {

    spherewidth [i] = (64 * 6.28) - sqrt (1024 - (abs (32 - i)) * (abs (32 - i))) * 6.28;

    tx [i] = (-1.57 + acos (((float) i - 32) / 32)) / (2 * PI);
  }


}

/*****************************************************************************/


void paint_frame_sphere (int mode, int t) {

  Window rootwin;
  Window dummywin;
  XImage *scr_shot;

  int x_start, y_start;

  int dummy;
  int mask_return = 0xFFFFFFFF;

  int dist;
  int sphere_frame;
  int x, y;
  float sine, cosine, scale;
  float src_x, src_y;
  int new_x, new_y;
  X_DATA_FORMAT * data;
  X_DATA_FORMAT * new_buf, *tex_buf;

  sphere_frame = t % 96;

  new_buf = (int *) malloc (192 * 128 * sizeof (X_DATA_FORMAT));
  tex_buf = (int *) malloc (96 * 64 * sizeof (X_DATA_FORMAT));

  rootwin = XRootWindow(display, 0); 

  XQueryPointer (display, rootwin, &dummywin, &dummywin, &x_start, &y_start, 
		 &dummy, &dummy, &mask_return);


  sine = cos (((float) (x_start +  y_start)) / 20.0);
  cosine = sin (((float) (x_start +  y_start)) / 20.0);

  x_start -= 96;
  y_start -= 64;

  if (x_start < 0) x_start = 0;
  if (y_start < 0) y_start = 0;

  if (x_start + 192 > screen_width) x_start = screen_width - 192;
  if (y_start + 128 > screen_height) y_start = screen_height - 128;

  scr_shot = XGetImage(display, rootwin, x_start, y_start, 192, 128, AllPlanes, ZPixmap);
  
  data = scr_shot->data;

  memcpy (new_buf, data, 192 * 128 * sizeof (X_DATA_FORMAT));

  for (y = 0; y < 64; y ++) {
    
    for (x = 0; x < 96; x ++) {

      
      scale = 1;

      switch (mode) {

      case 0:
      
	scale = 1 + 0.8 * cos (((float) ((x_start + y_start) * SCALE_POSITION_AMT + (abs (x - X_CENTER * 1.5) + abs (y - Y_CENTER)) * SCALE_DISTANCE_AMT + t * SPEED)) / 20.0);

	break;

      case 1:
	scale = 1 + 0.8 * cos (((float) ((x_start + y_start) * SCALE_POSITION_AMT + (abs (x - X_CENTER * 1.5) + abs (y - Y_CENTER)) * SCALE_DISTANCE_AMT)) / 20.0);

	break;

      case 2:
	scale = 1 + 0.8 * cos (((float) ((x_start + y_start) * SCALE_POSITION_AMT + (2 * abs (y - Y_CENTER)) * SCALE_DISTANCE_AMT + t * SPEED)) / 20.0);
	break;

      case 3:
	scale = 1 + 0.8 * cos (((float) ((x_start + y_start) * SCALE_POSITION_AMT + (2 * abs (x - X_CENTER * 1.5)) * SCALE_DISTANCE_AMT + t * SPEED)) / 20.0);
	break;

      case 4:
	scale = 1 + 0.8 * cos (((float) ((x_start + y_start) * SCALE_POSITION_AMT + (abs (x) + abs (y)) * SCALE_DISTANCE_AMT + t * SPEED)) / 20.0);

	break;

      case 5:
	scale = 2 + 0.8 * cos (((float) ((x_start + y_start) * SCALE_POSITION_AMT + (abs (x - X_CENTER * 1.5) + abs (y - Y_CENTER)) * SCALE_DISTANCE_AMT + t * SPEED)) / 20.0);
	break;

      case 6:
	scale = 0.3 + 0.2 * cos (((float) ((x_start + y_start) * SCALE_POSITION_AMT + (abs (x - X_CENTER * 1.5) + abs (y - Y_CENTER)) * SCALE_DISTANCE_AMT + t * SPEED)) / 20.0);
	
	break;
      }

      src_x = (sine * (float) x - cosine * (float) y) * scale;
      src_y = (cosine * (float) x + sine * (float) y) * scale;

      new_x = src_x + 32;
      new_y = src_y + 32;

      new_x = (new_x + 256) % 192;
      new_y = (new_y + 256) & 127;

      tex_buf [x + y * 96] = new_buf [new_x + (new_y * 192)];
    }
  }

  for (y = 0; y < 64; y ++) {
    for (x = 0; x < 64; x ++) {

      dist = ((x - 32) * (x - 32) + (y - 32) * (y - 32)) >> 4;

      if (dist < 42) {

	data [x + y * 192] = tex_buf [((int)(tx [x] * spherewidth [y] + sphere_frame) % 96) + y * 96];
      } else {
	data [x + y * 192] = 0xff000000;
      }
    }
  }


  XPutImage (display, win, NormalGC, scr_shot, 0, 0, 0, 0, 64, 64);

  XDestroyImage (scr_shot);
  free (new_buf);
  free (tex_buf);
}

/*****************************************************************************/

void paint_frame (int mode, int t) {


  Window rootwin;
  Window dummywin;
  XImage *scr_shot;

  int x_start, y_start;

  int dummy;
  int mask_return = 0xFFFFFFFF;

  int x, y;
  float sine, cosine, scale;
  float src_x, src_y;
  int new_x, new_y;
  X_DATA_FORMAT * data;
  X_DATA_FORMAT * new_buf;

  new_buf = (int *) malloc (128 * 128 * sizeof (X_DATA_FORMAT));
  

  rootwin = XRootWindow(display, 0); 

  XQueryPointer (display, rootwin, &dummywin, &dummywin, &x_start, &y_start, 
		 &dummy, &dummy, &mask_return);


  sine = cos (((float) (x_start +  y_start)) / 20.0);
  cosine = sin (((float) (x_start +  y_start)) / 20.0);

  x_start -= 64;
  y_start -= 64;

  if (x_start < 0) x_start = 0;
  if (y_start < 0) y_start = 0;

  if (x_start + 128 > screen_width) x_start = screen_width - 128;
  if (y_start + 128 > screen_height) y_start = screen_height - 128;

  scr_shot = XGetImage(display, rootwin, x_start, y_start, 128, 128, AllPlanes, ZPixmap);
  
  data = scr_shot->data;

  memcpy (new_buf, data, 128 * 128 * sizeof (X_DATA_FORMAT));


  for (y = 0; y < 64; y ++) {
    
    for (x = 0; x < 64; x ++) {
      
      scale = 1;

      switch (mode) {

      case 0:
      
      scale = 1 + 0.8 * cos (((float) ((x_start + y_start) * SCALE_POSITION_AMT + (abs (x - X_CENTER) + abs (y - Y_CENTER)) * SCALE_DISTANCE_AMT + t * SPEED)) / 20.0);

      break;

      case 1:
      scale = 1 + 0.8 * cos (((float) ((x_start + y_start) * SCALE_POSITION_AMT + (abs (x - X_CENTER) + abs (y - Y_CENTER)) * SCALE_DISTANCE_AMT)) / 20.0);

      break;

      case 2:
	scale = 1 + 0.8 * cos (((float) ((x_start + y_start) * SCALE_POSITION_AMT + (2 * abs (y - Y_CENTER)) * SCALE_DISTANCE_AMT + t * SPEED)) / 20.0);
      break;

      case 3:
	scale = 1 + 0.8 * cos (((float) ((x_start + y_start) * SCALE_POSITION_AMT + (2 * abs (x - X_CENTER)) * SCALE_DISTANCE_AMT + t * SPEED)) / 20.0);
      break;

      case 4:
	scale = 1 + 0.8 * cos (((float) ((x_start + y_start) * SCALE_POSITION_AMT + (abs (x) + abs (y)) * SCALE_DISTANCE_AMT + t * SPEED)) / 20.0);

      break;

      case 5:
	scale = 2 + 0.8 * cos (((float) ((x_start + y_start) * SCALE_POSITION_AMT + (abs (x - X_CENTER) + abs (y - Y_CENTER)) * SCALE_DISTANCE_AMT + t * SPEED)) / 20.0);
	break;
      case 6:
	scale = 0.3 + 0.2 * cos (((float) ((x_start + y_start) * SCALE_POSITION_AMT + (abs (x - X_CENTER) + abs (y - Y_CENTER)) * SCALE_DISTANCE_AMT + t * SPEED)) / 20.0);
	break;
      }

      src_x = (sine * (float) x - cosine * (float) y) * scale;
      src_y = (cosine * (float) x + sine * (float) y) * scale;

      new_x = src_x + 32;
      new_y = src_y + 32;

      new_x = (new_x + 256) & 127;
      new_y = (new_y + 256) & 127;

      data [x + y * 128] = new_buf [new_x + (new_y * 128)];
    }
  }

  XPutImage (display, win, NormalGC, scr_shot, 0, 0, 0, 0, 64, 64);

  XDestroyImage (scr_shot);
  free (new_buf);

}

/*****************************************************************************/

int main (int argc, char *argv[]) {
  XEvent Event;

  char wmrotozoom_mask_bits[64*64];
  int wmrotozoom_mask_width = 64;
  int wmrotozoom_mask_height = 64;

  double speed = 0.05;

  int sphere_mode = 0;

  int last_frame = 0, frame = 1, sphere_frame = 0;

  int i;

  int mode = 0;

  char *arg = NULL;

  createXBMfromXPM (wmrotozoom_mask_bits, wmrotozoom_master_xpm, wmrotozoom_mask_width, wmrotozoom_mask_height);
  openXwindow (argc, argv, wmrotozoom_master_xpm, wmrotozoom_mask_bits, wmrotozoom_mask_width, wmrotozoom_mask_height);

 
  AddMouseRegion (0, 0, 0, 64, 47);


  srand (clock ());

  for (i = 1; i < argc; i ++) {
    arg = argv[i];
    
    if (*arg=='-') {
      switch (arg[1]) {
	
      case 'v' :
	printf ("%s\n", VERSION);
	exit (0);
        break;
       
      case 's':

	sphere_mode = 1;
	break;
      }
    }
  }

  init_sphere ();
  
  screen_width = XDisplayWidth (display, 0);
  screen_height =  XDisplayHeight (display, 0);

  while (1) {

    if (speed > 0) {

      if (sphere_mode) {

	paint_frame_sphere (mode, sphere_frame);
	sphere_frame ++;

      } else {

	paint_frame (mode, frame);
      }

      last_frame = frame;
      frame ++;

      //RedrawWindow ();

    } else {
      usleep (USLEEP);
    }

    while (XPending (display)) {
      XNextEvent (display, &Event);
      switch (Event.type) {

      case Expose:

	break;

      case DestroyNotify:

	XCloseDisplay (display);
	exit (0);
	break;

      case ButtonPress:
	i = CheckMouseRegion (Event.xbutton.x, Event.xbutton.y);

	switch (i)
	  {

	  case 0:
	    mode = (mode + 1) % 8;

	    break;
	  }
	break;

      case ButtonRelease:
	i = CheckMouseRegion (Event.xbutton.x, Event.xbutton.y);

	break;
      }
    }
    usleep (USLEEP);
  }

  return 0;
}
